#====================== BEGIN GPL LICENSE BLOCK ======================
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
#======================= END GPL LICENSE BLOCK ========================

import bpy

LAYERS = [x == 0 for x in range(0,32)]


def local_matrix_with_constraints(pose_bone):
    """ Returns the local matrix of the given pose bone, with constraints
        applied.
    """
    mat = pose_bone.matrix.copy()
    rest = pose_bone.bone.matrix_local.copy()
    rest_inv = rest.copy()
    rest_inv.invert()
    
    if pose_bone.parent:
        mat2_inv = pose_bone.parent.matrix.copy()
        mat2_inv.invert()
        rest2 = pose_bone.parent.bone.matrix_local.copy()
        
        mat = rest_inv * (rest2 * (mat2_inv * mat))
    else:
        mat = rest_inv * mat
    
    return mat


def make_deform_only_rig(context, obj, name):
    """ Creates an armature object with only the deforming
        bones of the given armature object.
    """
    obj_d = bpy.data.objects.new(name, obj.data.copy())
    context.scene.objects.link(obj_d)

    context.scene.objects.active = obj_d
    obj_d.select = True

    for ob in context.scene.objects:
        obj.select = False

    bpy.ops.object.mode_set(mode='EDIT')

    for bone in obj_d.data.edit_bones:
        if not bone.use_deform:
            obj_d.data.edit_bones.remove(bone)
        else:
            bone.layers = LAYERS

    bpy.ops.object.mode_set(mode='OBJECT')

    for bone in obj_d.pose.bones:
        bone.rotation_mode = 'QUATERNION'

    obj_d.data.layers = LAYERS

    return obj_d
    


class MakeDeformOnlyRig(bpy.types.Operator):
    """ Generates a rig from the active metarig armature
    """
    bl_idname = "pose.make_deform_only_rig"
    bl_label = "Make Deform Only Rig"

    def execute(self, context):
        make_deform_only_rig(context, context.active_object, "rig_deform_only")

        return {'FINISHED'}


def bake_deform_action(context, obj):
    """ Bakes the action on the current armature object to only the deform
        bone's motions.
    """
    if obj.animation_data:
        if obj.animation_data.action:
            # Create a deform-only rig
            rigd = make_deform_only_rig(context, obj, "temp_def_rig")

            # Constrain the deform-only rig to the main rig
            con = rigd.constraints.new('COPY_TRANSFORMS')
            con.name = "copy_transforms"
            con.target = obj
            for bone in rigd.pose.bones:
                con = bone.constraints.new('COPY_TRANSFORMS')
                con.name = "copy_transforms"
                con.target = obj
                con.subtarget = bone.name
            
            # Create a new action
            action = bpy.data.actions.new(name=obj.animation_data.action.name + "_deform_only")

            # Create the fcurves for each of the deform bones
            fcurves = {}
            for bone in rigd.pose.bones:
                locx = action.fcurves.new(data_path="pose.bones['%s'].location" % bone.name, array_index=0, action_group=bone.name)
                locy = action.fcurves.new(data_path="pose.bones['%s'].location" % bone.name, array_index=1, action_group=bone.name)
                locz = action.fcurves.new(data_path="pose.bones['%s'].location" % bone.name, array_index=2, action_group=bone.name)

                rotw = action.fcurves.new(data_path="pose.bones['%s'].rotation_quaternion" % bone.name, array_index=0, action_group=bone.name)
                rotx = action.fcurves.new(data_path="pose.bones['%s'].rotation_quaternion" % bone.name, array_index=1, action_group=bone.name)
                roty = action.fcurves.new(data_path="pose.bones['%s'].rotation_quaternion" % bone.name, array_index=2, action_group=bone.name)
                rotz = action.fcurves.new(data_path="pose.bones['%s'].rotation_quaternion" % bone.name, array_index=3, action_group=bone.name)

                sclx = action.fcurves.new(data_path="pose.bones['%s'].scale" % bone.name, array_index=0, action_group=bone.name)
                scly = action.fcurves.new(data_path="pose.bones['%s'].scale" % bone.name, array_index=1, action_group=bone.name)
                sclz = action.fcurves.new(data_path="pose.bones['%s'].scale" % bone.name, array_index=2, action_group=bone.name)

                fcurves[bone.name] = [locx, locy, locz, rotw, rotx, roty, rotz, sclx, scly, sclz]

            # Bake the deformation action
            f_backup = context.scene.frame_current
            for f in range(context.scene.frame_start, context.scene.frame_end + 1):
                context.scene.frame_set(f)

                for bone in rigd.pose.bones:
                    m = local_matrix_with_constraints(bone)
                    loc = m.translation_part()
                    rot = m.rotation_part().to_quat()
                    scl = m.scale_part()

                    values = [loc[0], loc[1], loc[2], rot[0], rot[1], rot[2], rot[3], scl[0], scl[1], scl[2]]
                    for (fcurve, value) in zip(fcurves[bone.name], values):
                        fcurve.keyframe_points.add(f, value)
            context.scene.frame_set(f_backup)

            # Make it so that the action won't disappear upon saving/reloading
            action.use_fake_user = True

            # Delete the deform-only rig
            context.scene.objects.unlink(rigd)
            bpy.data.objects.remove(rigd)

            
                


class BakeDeformAction(bpy.types.Operator):
    """ Generates a rig from the active metarig armature
    """
    bl_idname = "pose.bake_deform_action"
    bl_label = "Bake Deform Action"

    def execute(self, context):
        bake_deform_action(context, context.active_object)

        return {'FINISHED'}

